#ifndef IWAH_H
#define IWAH_H

#include <inttypes.h>

#define GCB 	0	//Common Dunlop GCB-95
#define VOX 	1 	//Vox V847 voicing
#define DUNLOP	2	//"The Original" Crybaby
#define MCCOY	3 	//Clyde McCoy voicing
#define VOCAL	4   //Vox with a little extra Q and some low-end brought up a bit
#define EXTREME 5 	//crazy synth-like
#define MAX_WAHS 6  //Update as wahs are added

typedef struct iwah_t {
    //Circuit parameters
    //Using these makes it straight-forward to model other
    //variants of the circuit
    float Lp; //RLC tank inductor
    float Cf; //feedback capacitor
    float Ci; //input capacitor
    float re; //equivalent resistance looking into input BJT base
    float Rp; //resistor placed parallel with the inductor
    float Ri; //input feed resistor
    float Rs; //RLC tank to BJT base resistor (dry mix)

    //Gain-setting components
    float Rc; //BJT gain stage collector resistor
    float Rpot; //Pot resistance value
    float Rbias; //Typically 470k bias resistor shows up in parallel with output
    float Re; //BJT gain stage emitter resistor
    float gf; //forward gain of BJT amplifier

    //High-Pass biquad coefficients
    float b0h, b1h, b2h;
    float a0h, a1h, a2h;

    //Band-Pass biquad coefficients
    float b0b, b2b;
    float a0b, a1b, a2b;

    //Final combined biquad coefficients used by run_filter()
    float b0;
    float b1;
    float b2;
    float a0, a0c;
    float a1, a1c;
    float a2, a2c;
    float ax; //this is used to store 1/a0 to avoid several 1/x operations

    //First order high-pass filter coefficients
    //y[n] = ghpf * ( x[n] - x[n-1] ) - a1p*y[n-1]
    float a1p;
    float ghpf;

    //biquad state variables
    float y1;
    float y2;
    float x1;
    float x2;

    //First order high-pass filter state variables
    float yh1;
    float xh1;

    //Bypass the effect
    bool bypass;

} iwah_coeffs;

typedef struct iwah_fixed_t {

    //Fixed-point implementation
    int32_t nb[3];  //b0, b1, b2 for fixed-point implementation
    int32_t na[3]; //a0, a1, a2
    int32_t nac[3]; //a0c, a1c, a2c
    int32_t nax;

    //First order high-pass filter coefficients
    //y[n] = ghpf * ( x[n] - x[n-1] ) - a1p*y[n-1]
    int32_t na1p;
    int32_t nghpf;

    //biquad state variables
    int32_t ny[2];
    int32_t nx[2];

    //First order high-pass filter state variables
    int32_t nyh1, nxh1;

} iwah_coeffs_fixed;

//Initialize the struct and allocate memory
iwah_coeffs* make_iwah(iwah_coeffs*, float);
void make_iwah_fixed_point(iwah_coeffs_fixed* fpcf, iwah_coeffs* cf);
//(float x, float gp, iwah_coeffs* cf)
// input sample, pot gain (0...1), iwah struct
float iwah_tick(float, float, iwah_coeffs*);
void iwah_tick_n(iwah_coeffs*, float*, float*, int );
void iwah_tick_n_fixed_point(iwah_coeffs_fixed*, int32_t*, int32_t*, uint32_t);

//Select preset circuit voicings
void iwah_circuit_preset(int , iwah_coeffs*, float );

//Bypass effect
void iwah_bypass(iwah_coeffs*, bool);


// Multiply two S.31 fractional integers, and return the 32 most significant
// bits after a shift left by the constant z.
// This comes from rockbox.org


//
// FRACMUL implementations from Rockbox (rockbox.org) fracmul.h
// Of particular interest will be the ARM ASM implementation.
//
        /* Rockbox ASM implementations for the ARM specification */
    // /* Multiply two S.31 fractional integers and return the sign bit and the
    //  * 31 most significant bits of the result.
    //  */
    // static inline int32_t FRACMUL(int32_t x, int32_t y)
    // {
    //     int32_t t, t2;
    //     asm ("smull    %[t], %[t2], %[a], %[b]\n\t"
    //          "mov      %[t2], %[t2], asl #1\n\t"
    //          "orr      %[t], %[t2], %[t], lsr #31\n\t"
    //          : [t] "=&r" (t), [t2] "=&r" (t2)
    //          : [a] "r" (x), [b] "r" (y));
    //     return t;
    // }
    //
    // /* Multiply two S.31 fractional integers, and return the 32 most significant
    //  * bits after a shift left by the constant z.
    //  */
    // static FORCE_INLINE int32_t FRACMUL_SHL(int32_t x, int32_t y, int z)
    // {
    //     int32_t t, t2;
    //     asm ("smull    %[t], %[t2], %[a], %[b]\n\t"
    //          "mov      %[t2], %[t2], asl %[c]\n\t"
    //          "orr      %[t], %[t2], %[t], lsr %[d]\n\t"
    //          : [t] "=&r" (t), [t2] "=&r" (t2)
    //          : [a] "r" (x), [b] "r" (y),
    //            [c] "Mr" ((z) + 1), [d] "Mr" (31 - (z)));
    //     return t;
    // }

    /* Rockbox generic implementations */
static inline int32_t FRACMUL(int32_t x, int32_t y)
{
    return (int32_t) (((int64_t)x * y) >> 31);
}

static inline int32_t FRACMUL_SHL(int32_t x, int32_t y, int z)
{
    return (int32_t) (((int64_t)x * y) >> (31 - z));
}

#endif //IWAH_H
